@@ -78,43 +78,54 @@ def is_nodejs_runtime(lambda_details):
     return runtime.startswith('nodejs')
 
 
+def _store_logs(func_details, log_output, invocation_time=None, container_id=None):
+    log_group_name = '/aws/lambda/%s' % func_details.name()
+    container_id = container_id or short_uid()
+    invocation_time = invocation_time or int(time.time() * 1000)
+    invocation_time_secs = int(invocation_time / 1000)
+    time_str = time.strftime('%Y/%m/%d', time.gmtime(invocation_time_secs))
+    log_stream_name = '%s/[LATEST]%s' % (time_str, container_id)
+    return store_cloudwatch_logs(log_group_name, log_stream_name, log_output, invocation_time)
+
+
 class LambdaExecutor(object):
     """ Base class for Lambda executors. Subclasses must overwrite the _execute method """
-
     def __init__(self):
         # keeps track of each function arn and the last time it was invoked
         self.function_invoke_times = {}
 
-    def execute(self, func_arn, func_details, event, context=None, version=None, asynchronous=False):
-
+    def execute(self, func_arn, func_details, event, context=None, version=None,
+            asynchronous=False, callback=None):
         def do_execute(*args):
             # set the invocation time in milliseconds
             invocation_time = int(time.time() * 1000)
             # start the execution
+            raised_error = None
+            result = None
+            dlq_sent = None
             try:
-                result, log_output = self._execute(func_arn, func_details, event, context, version)
+                result = self._execute(func_arn, func_details, event, context, version)
             except Exception as e:
+                raised_error = e
                 if asynchronous:
                     if get_from_event(event, 'eventSource') == EVENT_SOURCE_SQS:
                         sqs_queue_arn = get_from_event(event, 'eventSourceARN')
                         if sqs_queue_arn:
                             # event source is SQS, send event back to dead letter queue
-                            sqs_error_to_dead_letter_queue(sqs_queue_arn, event, e)
+                            dlq_sent = sqs_error_to_dead_letter_queue(sqs_queue_arn, event, e)
                     else:
-                        # event source is not SQS, send back to lambda dead letter
+                        # event source is not SQS, send back to lambda dead letter queue
                         lambda_error_to_dead_letter_queue(func_details, event, e)
                 raise e
             finally:
                 self.function_invoke_times[func_arn] = invocation_time
-            # forward log output to cloudwatch logs
-            self._store_logs(func_details, log_output, invocation_time)
+                callback and callback(result, func_arn, event, error=raised_error, dlq_sent=dlq_sent)
             # return final result
-            return result, log_output
+            return result
 
         # Inform users about asynchronous mode of the lambda execution.
         if asynchronous:
-            LOG.debug('Lambda executed in Event (asynchronous) mode, no response from this '
-                      'function will be returned to caller')
+            LOG.debug('Lambda executed in Event (asynchronous) mode, no response will be returned to caller')
             FuncThread(do_execute).start()
             return None, 'Lambda executed asynchronously.'
 
@@ -130,16 +141,9 @@ class LambdaExecutor(object):
     def cleanup(self, arn=None):
         pass
 
-    def _store_logs(self, func_details, log_output, invocation_time):
-        log_group_name = '/aws/lambda/%s' % func_details.name()
-        invocation_time_secs = int(invocation_time / 1000)
-        time_str = time.strftime('%Y/%m/%d', time.gmtime(invocation_time_secs))
-        log_stream_name = '%s/[$LATEST]%s' % (time_str, short_uid())
-        return store_cloudwatch_logs(log_group_name, log_stream_name, log_output, invocation_time)
-
-    def run_lambda_executor(self, cmd, event=None, env_vars={}):
-        process = run(cmd, asynchronous=True, stderr=subprocess.PIPE, outfile=subprocess.PIPE, env_vars=env_vars,
-                      stdin=True)
+    def run_lambda_executor(self, cmd, event=None, func_details=None, env_vars={}):
+        process = run(cmd, asynchronous=True, stderr=subprocess.PIPE, outfile=subprocess.PIPE,
+                      env_vars=env_vars, stdin=True)
         result, log_output = process.communicate(input=event)
         try:
             result = to_str(result).strip()
@@ -154,11 +158,18 @@ class LambdaExecutor(object):
             additional_logs, _, result = result.rpartition('\n')
             log_output += '\n%s' % additional_logs
 
+        log_formatted = log_output.strip().replace('\n', '\n> ')
+        func_arn = func_details and func_details.arn()
+        LOG.debug('Lambda %s result / log output:\n%s\n> %s' % (func_arn, result.strip(), log_formatted))
+
+        # store log output - TODO get live logs from `process` above?
+        _store_logs(func_details, log_output)
+
         if return_code != 0:
             raise Exception('Lambda process returned error status code: %s. Result: %s. Output:\n%s' %
                 (return_code, result, log_output))
 
-        return result, log_output
+        return result
 
 
 class ContainerInfo:
@@ -187,7 +198,6 @@ class LambdaExecutorContainers(LambdaExecutor):
         return None
 
     def _execute(self, func_arn, func_details, event, context=None, version=None):
-
         lambda_cwd = func_details.cwd
         runtime = func_details.runtime
         handler = func_details.handler
@@ -243,15 +253,13 @@ class LambdaExecutorContainers(LambdaExecutor):
 
         # lambci writes the Lambda result to stdout and logs to stderr, fetch it from there!
         LOG.info('Running lambda cmd: %s' % cmd)
-        result, log_output = self.run_lambda_executor(cmd, stdin, environment)
-        log_formatted = log_output.strip().replace('\n', '\n> ')
-        LOG.debug('Lambda %s result / log output:\n%s\n>%s' % (func_arn, result.strip(), log_formatted))
-        return result, log_output
+        result = self.run_lambda_executor(cmd, stdin, env_vars=environment, func_details=func_details)
+
+        return result
 
 
 class LambdaExecutorReuseContainers(LambdaExecutorContainers):
     """ Executor class for executing Lambda functions in re-usable Docker containers """
-
     def __init__(self):
         super(LambdaExecutorReuseContainers, self).__init__()
         # locking thread for creation/destruction of docker containers.
@@ -266,7 +274,6 @@ class LambdaExecutorReuseContainers(LambdaExecutorContainers):
         self.port_offset = LAMBDA_SERVER_PORT_OFFSET
 
     def prepare_execution(self, func_arn, env_vars, runtime, command, handler, lambda_cwd):
-
         # check whether the Lambda has been invoked before
         has_been_invoked_before = func_arn in self.function_invoke_times
 
@@ -514,11 +521,8 @@ class LambdaExecutorReuseContainers(LambdaExecutorContainers):
         :param func_arn: The ARN of the lambda function.
         :return: name of the container network
         """
-
         with self.docker_container_lock:
-
             status = self.get_docker_container_status(func_arn)
-
             # container does not exist
             if status == 0:
                 return ''
@@ -578,7 +582,6 @@ class LambdaExecutorReuseContainers(LambdaExecutorContainers):
 
 
 class LambdaExecutorSeparateContainers(LambdaExecutorContainers):
-
     def __init__(self):
         super(LambdaExecutorSeparateContainers, self).__init__()
         self.next_port = 1
@@ -586,7 +589,6 @@ class LambdaExecutorSeparateContainers(LambdaExecutorContainers):
         self.port_offset = LAMBDA_API_PORT_OFFSET
 
     def prepare_event(self, environment, event_body):
-
         # Tell Lambci to use STDIN for the event
         environment['DOCKER_LAMBDA_USE_STDIN'] = '1'
         return event_body.encode()
@@ -644,7 +646,6 @@ class LambdaExecutorSeparateContainers(LambdaExecutorContainers):
 
 
 class LambdaExecutorLocal(LambdaExecutor):
-
     def _execute(self, func_arn, func_details, event, context=None, version=None):
         lambda_cwd = func_details.cwd
         environment = func_details.envvars.copy()
@@ -672,9 +673,14 @@ class LambdaExecutorLocal(LambdaExecutor):
         for stream in (c.stdout(), c.stderr()):
             if stream:
                 log_output += ('\n' if log_output else '') + stream
-        return result, log_output
 
-    def execute_java_lambda(self, event, context, handler, main_file):
+        # store logs to CloudWatch
+        _store_logs(func_details, log_output)
+
+        return result
+
+    def execute_java_lambda(self, event, context, main_file, func_details=None):
+        handler = func_details.handler
         opts = config.LAMBDA_JAVA_OPTS if config.LAMBDA_JAVA_OPTS else ''
         event_file = EVENT_FILE_PATTERN.replace('*', short_uid())
         save_file(event_file, json.dumps(event))
@@ -683,10 +689,8 @@ class LambdaExecutorLocal(LambdaExecutor):
         classpath = '%s:%s:%s' % (LAMBDA_EXECUTOR_JAR, main_file, Util.get_java_classpath(main_file))
         cmd = 'java %s -cp %s %s %s %s' % (opts, classpath, LAMBDA_EXECUTOR_CLASS, class_name, event_file)
         LOG.warning(cmd)
-        result, log_output = self.run_lambda_executor(cmd)
-        LOG.debug('Lambda result / log output:\n%s\n> %s' % (
-            result.strip(), log_output.strip().replace('\n', '\n> ')))
-        return result, log_output
+        result = self.run_lambda_executor(cmd, func_details=func_details)
+        return result
 
 
 class Util:
@@ -695,10 +699,17 @@ class Util:
     @classmethod
     def get_java_opts(cls):
         opts = config.LAMBDA_JAVA_OPTS or ''
+        # Replace _debug_port_ with a random free port
         if '_debug_port_' in opts:
             if not cls.debug_java_port:
                 cls.debug_java_port = get_free_tcp_port()
             opts = opts.replace('_debug_port_', ('%s' % cls.debug_java_port))
+        else:
+            # Parse the debug port from opts
+            m = re.match('.*address=(\\d+).*', opts)
+            if m is not None:
+                cls.debug_java_port = m.groups()[0]
+
         return opts
 
     @classmethod
